#ifndef HEAP_HPP_
#define HEAP_HPP_
#define heap_hpp_version "$Revision: 1.4 $"

#include "Assertion.hpp"
#include "BitVector.hpp"
#include	"Random.hpp"
#include <vector>
#include "Logging.hpp"



template<class KeyType, class ValueType>
struct ValCompPrefMin{
	bool better(const KeyType& ka, const ValueType& a, const KeyType& kb, const ValueType& b)const{return a<b;}
};

template<class KeyType, class ValueType>
struct ValCompPrefMax{
	bool better(const KeyType& ka, const ValueType& a, const KeyType& kb, const ValueType& b)const{return a>b;}
};




/**
 * Heap structure that ranks all elements according to some value.*/
template<class KeyType, class ValueType, class ValueCompare = ValCompPrefMin<KeyType,ValueType>, class KeyTypeConverter = DfltKeyTypeConverter<KeyType> >
class Heap{
	static inline uint32_t   parent		(uint32_t index)	 	{return (index-1) >> 1;};
	static inline uint32_t   leftChild	 (uint32_t index)		{return (index << 1) + 1;};
	static inline uint32_t   rightChild	(uint32_t index)	 	{return (index << 1) + 2;};
	static const uint32_t  	IVLD_IDX = ~0;

protected:
	const KeyTypeConverter& 	converter;
	const ValueCompare&		comp;

	/**Internal data-structure to store the information for one element.*/
	struct ElementInfo{
		uint32_t 	index_in_vec;
		ValueType 	value;

		/**Constructs an invalid emement.*/
		inline ElementInfo() :index_in_vec(IVLD_IDX),value(0){};
		inline ElementInfo(uint32_t idx,const ValueType& v):index_in_vec(idx),value(v){}
		inline bool contained()const {return index_in_vec != IVLD_IDX;}
	};


	/**Holds an ElementInfo for each element in the heap.*/
	std::vector<ElementInfo > 	e_info;

	inline uint32_t	moveUp(const uint32_t element_id, uint32_t heap_index);
	inline uint32_t	moveDown(const uint32_t element_id, uint32_t heap_index);
	inline void		updateAtPos(uint32_t pos);

protected:
	/**The actual heap. Each element of the heap vector is an index of the original element into e_info.*/
	std::vector<uint32_t> 			eles;

	inline			Heap(const Heap& h);

	/**No assignment.*/
	inline const	Heap<KeyType,ValueType,ValueCompare,KeyTypeConverter>& operator=(const Heap<KeyType,ValueType,ValueCompare,KeyTypeConverter>& h);

	/**Returns element id of the element that was at the specified position pos.
	 * Does not ensure the heap property*/
	inline uint32_t	lazyRemoveAtPos(uint32_t pos);

	inline bool		checkHeapProperty(uint32_t heap_index) const;

	template<class PV>
	inline void		printHeap(PV& val_printer, uint32_t heap_index, uint32_t levl,uint32_t parnt) const;

	inline bool		better(uint32_t el_a, uint32_t el_b) const
					{return comp.better(converter.fromInt(el_a),e_info[el_a].value,converter.fromInt(el_b),e_info[el_b].value);};

	inline bool		betterOrEq(uint32_t el_a, uint32_t el_b) const
					{return !comp.better(converter.fromInt(el_b),e_info[el_b].value,converter.fromInt(el_a),e_info[el_a].value);};


public:

	Heap(const ValueCompare& compare=ValueCompare(),const KeyTypeConverter& conv = KeyTypeConverter())
					:converter(conv),comp(compare){}

	/**Reserve enough space for the specified number of elements*/
	void			reserve(uint32_t capacity){
					if(capacity >= e_info.size()){
						e_info	.reserve(capacity);
						eles	.reserve(capacity);}}


	/**Validates the given element on the heap.
	 * If not yet contained, the initial value is set to default_val*/
	inline void		insertElementsTo(const uint32_t& req_sz, const ValueType& default_val);

	/**Ensures to have values for all elements up to max_element. New elements are not inserted into the heap*/
	inline void		ensureElementValsTo(const uint32_t& req_sz, const ValueType& default_val);

	/**Reactivates an element that was previously removed by the function removeTop or removeAtPos.*/
	inline void		restoreElementIf(const KeyType& element);

	/**Return the current rank value of the specified element*/
	ValueType 		getValue(const KeyType& element) const{
					asrt(converter.toInt(element) < e_info.size()); return e_info[converter.toInt(element)].value;}

	ValueType		getMax() const{
					return eles.empty() ? ValueType(0) : e_info[eles[0]].value;}

	void			setValue(const KeyType& e, const ValueType& v){
					asrt(converter.toInt(e) < e_info.size()); sanity(checkHeapProperty(0));
					e_info[converter.toInt(e)].value = v;
					if(e_info[converter.toInt(e)].contained()){
						updateAtPos(e_info[converter.toInt(e)].index_in_vec);}
						sanity(checkHeapProperty(0));}

	ValueType		getTop() const{
					return e_info[eles[0]].value;}


	/**Checks if the heap is empty*/
	bool		 	empty()	{return eles.empty();}
	uint32_t 		size()	{return eles.size();}

	void			clear(){
					for(uint32_t i=eles.size();i--;){e_info[eles[i]].index_in_vec=IVLD_IDX;}
					eles.clear();}

	bool			contains(const KeyType& e)const{
					asrt(converter.toInt(e) < e_info.size());
					return e_info[converter.toInt(e)].contained();}

	KeyType			getAtPos(uint32_t pos){
					asrt(pos < eles.size());
					return converter.fromInt(eles[pos]);}

	KeyType			removeAtPos(uint32_t pos){
					asrt(pos < eles.size());
					uint32_t ret_idx 	=	lazyRemoveAtPos(pos);
					if(pos < eles.size()){	updateAtPos(pos);}	sanity(checkHeapProperty(0));
					return converter.fromInt(ret_idx);}

	void			removeIfContained(const KeyType& e){
					asrt(converter.toInt(e) < e_info.size());
					ElementInfo& ef = e_info[converter.toInt(e)];
					if(!ef.contained()) return;
					removeAtPos(ef.index_in_vec);sanity(checkHeapProperty(0));}

	/**Get the element with the highest rank value and remove it from the heap.*/
	KeyType			removeTop(){
					asrt(!empty());
					uint32_t ret_idx = lazyRemoveAtPos(0);
					if(eles.size() > 1){moveDown(eles[0],0);}sanity(checkHeapProperty(0));
					return converter.fromInt(ret_idx);}

	KeyType			removeRandom(Random& randy){
					return 	removeAtPos(randy.nextInt(eles.size()));
	}

	bool			asrtHeapProperty() const {return checkHeapProperty(0);}
	template<class PV>
	void			printHeap(PV& val_printer) const {printHeap(val_printer, 0,0,IVLD_IDX);}
};



template<class KeyType, class ValueType, class ValueCompare,  class KeyTypeConverter = DfltKeyTypeConverter<KeyType> >
class BumpHeap: public Heap<KeyType,ValueType,ValueCompare,KeyTypeConverter >{
	typedef Heap<KeyType,ValueType,ValueCompare,KeyTypeConverter > super;
	const ValueType	max;
	ValueType	   	increment;
	const ValueType decay;
	const typename ValueType::denominator_t weaken;
public:
	inline bool		activityBump(const KeyType& element);
	void			decayAll()				{increment *= decay; };
	BumpHeap(ValueType MAX, ValueType dec,const ValueCompare& comp):
		super(comp,KeyTypeConverter()),
		max(MAX),
		increment(1, ValueType::minNonZeroNumber().getLogOfDenominator() - (2* dec.getLogOfDenominator())),
		decay(dec),
		weaken(increment.getLogOfDenominator()){}
};



//Inline Implementation


template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
uint32_t Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::moveUp(const uint32_t element_id, uint32_t heap_index){
	asrt(heap_index < eles.size()); asrt(eles[heap_index] == element_id);
	uint32_t parent_index;
	while( heap_index != 0
		&& (parent_index = parent(heap_index), better(element_id,eles[parent_index]))){

		eles[heap_index]								=   eles[parent_index];
		e_info[eles[heap_index]].index_in_vec			=   heap_index;
		asrt(e_info[eles[heap_index]].contained());
		heap_index										=   parent_index;
	}
	eles[heap_index]									=   element_id;
	e_info[element_id].index_in_vec						=   heap_index;
	asrt(e_info[element_id].contained());
	return heap_index;
}

template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
uint32_t Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::moveDown(const uint32_t element_id, uint32_t heap_index){
	asrt(heap_index < eles.size()); asrt(eles[heap_index] == element_id);
	uint32_t child_index;
	while((child_index = leftChild(heap_index), child_index < eles.size())){

		//get the child with the higher rank
		if(child_index+1 < eles.size() && better(eles[child_index+1],eles[child_index])){
			++child_index;} //the right one is bigger

		if(betterOrEq(element_id,eles[child_index]))	break;

		eles[heap_index]								=   eles[child_index];
		e_info[eles[heap_index]].index_in_vec			=   heap_index;
		asrt(e_info[eles[heap_index]].contained());
		heap_index										=   child_index;
	}
	eles[heap_index]									=   element_id;
	e_info[element_id].index_in_vec						=   heap_index;
	return heap_index;
}


template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::updateAtPos(uint32_t pos){
	asrt(pos < eles.size());
	pos = moveDown(eles[pos],pos); asrt(e_info[eles[pos]].index_in_vec==pos);
	moveUp(eles[pos],pos);	asrt(e_info[eles[pos]].index_in_vec==pos);
	sanity(checkHeapProperty(0));
}



//has to be ensured that not empty
template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
uint32_t Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::lazyRemoveAtPos(uint32_t pos){
	asrt(pos < eles.size()); asrt(!eles.empty());

	uint32_t e_idx					= eles[pos]; asrt(e_idx < e_info.size());
	uint32_t last					= eles.back(); eles.pop_back();
	eles[pos]						= last;
	e_info[last].index_in_vec		= pos;
	e_info[e_idx].index_in_vec		= IVLD_IDX;//we leave value unchanged for re-adding

	return e_idx;
}




template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::insertElementsTo(const uint32_t& req, const ValueType& default_val){

	if(e_info.size() > req) return;
	e_info					.reserve(req);
	eles					.reserve(req);

	for(uint32_t i = e_info.size(); i < req; ++i){
		asrt(i == e_info.size());
		e_info	.push_back(ElementInfo(i,default_val));
		eles	.push_back(i);
		moveUp(i,eles.size()-1);
	}
	sanity(checkHeapProperty(0));
}

template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::ensureElementValsTo(const uint32_t& req, const ValueType& default_val){
	if(e_info.size() > req) return;
	e_info.resize(req,ElementInfo(IVLD_IDX,default_val));
}


template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::restoreElementIf(const KeyType& element){
	asrt(e_info.size() > converter.toInt(element));

	ElementInfo& ef = e_info[converter.toInt(element)];
	if(ef.contained()) return;

	ef		.index_in_vec = eles.size();
	eles	.push_back(converter.toInt(element));
	moveUp(converter.toInt(element),ef.index_in_vec);
	sanity(checkHeapProperty(0));
}


template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
bool Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::checkHeapProperty(uint32_t i) const{
	return i >= eles.size()
		  || ((i == 0 || betterOrEq(eles[parent(i)],eles[i])) && checkHeapProperty(leftChild(i)) && checkHeapProperty(rightChild(i)));
}



template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
template<class PV>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::printHeap(PV& val_printer, uint32_t i,uint32_t levl,uint32_t parnt) const{
	if(i >= eles.size()) return;
	std::string in("c "); for(uint32_t k=levl;k--;){in +=" ";}
	cout << in << "Heap element " << eles[i] << " with value " << val_printer(e_info[eles[i]].value) << " at levl " << levl << " position " << i << " parent " <<parnt << endl;
	printHeap(val_printer,leftChild(i),levl+1,i);
	printHeap(val_printer, rightChild(i),levl+1,i);
}



template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
bool BumpHeap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::activityBump(const KeyType& element){

	typename super::ElementInfo& ef = super::e_info[super::converter.toInt(element)];
	bool ret = true;

	if((ef.value += increment) > max){
		increment	.weaken(weaken);
		int32_t i = super::e_info.size();
		while(i--){ super::e_info[i].value.weaken(weaken);}
		ret = false;
	}

	if(ef.contained()){
		super::moveUp(super::converter.toInt(element),ef.index_in_vec);}

	sanity(super::checkHeapProperty(0));
	return ret;
}






#endif /*DECAYHEAP_HPP_*/


